/*
 *				**** CSV FMT ****
 *
 * This program allows you to reformat CSV files
 * based on a format string
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "wrpcsv.h"
#include "csvfmt.h"

char *Prog = NULL;
char *no_data = "[nodata]";

struct csvfmt *create_csvfmt(void);
void free_csvfmt(struct csvfmt *fmt);
void csvfmt(FILE *fp, struct csvfmt *fmt, int nskip, char *fname);
struct csvfmt *process_fmt( char *fname);
void dump_fmt(struct csvfmt *fmt);
void append_fmt(struct csvfmt *ret, struct csvfmt *fmt);
void copy_file(FILE *fpin, FILE *fpout);

int main(int argc, char *argv[])
{
	char buf[BUFSIZ];
   struct csvfmt *fmt = NULL;
   FILE *fpout;
   char *p;
   char *bfile=NULL, *efile=NULL;
   FILE *bfp=NULL, *efp=NULL; /* beginning and ending files */
   int nskip = 0; /* number of lines to skip */

   fpout = stdout;

   Prog = argv[0];
   argc--; argv++;

   while (argc > 0)
   {
   	if (argv[0][0] == '-') switch(argv[0][1])
      { /* option */
      case 'b': /* beginning file */
      	if (argv[0][2] == '\0')
         { /* file name is following argument */
         	argc--; argv++;/* skip this argument */
            bfile = argv[0]; /* file name */
            if (bfile == NULL)
            {
            	fprintf(stderr, "Missing beginning file name\n");
               exit(3);
            }
         }
         else
         { /* file name right after -b flag */
         	bfile = argv[0]+2;
         }
         if ((bfp = fopen(bfile, "r")) == NULL)
         {
         	fprintf(stderr, "Error opening beginning file (%s)\n", bfile);
            exit(4);
         }
         break;

      case 'e': /* ending file */
      	if (argv[0][2] == '\0')
         { /* file name is  next argument */
         	argc--; argv++; /* skip this argument */
            efile = argv[0];
            if (efile == NULL)
            {
            	fprintf(stderr, "Missing ending file name\n");
               exit(5);
            }
         }
			else
         { /* file name follows -e flag */
         	efile = argv[0] + 2;
         }
         if ((efp = fopen(efile, "r")) == NULL)
         {
         	fprintf(stderr, "Unable to open ending file (%s)\n", efile);
            exit(6);
         }
      	break;

      case 'n': /* suppress no-data message */
      	no_data = "";
         break;

      case 'o': /* output file */
      	if (argv[0][2] == '\0')
         { /* file name is the following argument */
         	argc--; argv++;
            p = argv[0];	/* file name */
            if (p == NULL)
            {
            	fprintf(stderr, "File name missing after -o argument\n");
               exit(1);
            }
         }
         else
         { /* file name right after -o */
         	p = argv[0] + 2;
         }
      	if ((fpout = fopen(p, "w")) == NULL)
         {
         	fprintf(stderr, "Can't open output file (%s)\n", p);
            exit(2);
         }
         break;

      case 'f': /* output format file */
      	if (argv[0][2] == '\0')
         { /* format file is next argument */
         	argc--; argv++;
            p = argv[0];	/* file name */
         }
         else
         { /* file name immediately after -f */
         	p = argv[0] + 2;
         }
         fmt = process_fmt(p);
#if EBUG
         dump_fmt(fmt);
#endif
         break;

		case 's': /* number of lines to skip */
      	if (argv[0][2] == '\0')
         { /* number is next argument */
         	argc--; argv++;
            if (argc>=0) p = argv[0]; else p="";
         }
         else
         { /* number right after -s flag */
         	p = argv[0]+2;
         }
         nskip = atoi(p);
         break;

      default:
      	fprintf(stderr, "Invalid option (%s)\n", argv[0]);
      } /* end handling option */
      else
      { /* file name(s) */
      	if (bfp != NULL) copy_file(bfp, fpout);
      	csvfmt(fpout, fmt, nskip, argv[0]);
         if (efp != NULL) copy_file(efp, fpout);
      }
   	argc--; argv++;
   }

	return 0;
}

struct csvfmt *process_fmt(char *fname)
{ /* process format file and build structure */
	char buf[BUFSIZ];
   FILE *fp;
   struct csvfmt *ret = NULL, *fmt, *fmtnew;
   char *p, *q;
   char c;
#if EBUG
   printf("process_fmt, fname='%s'\n", fname);
#endif
   if ((fp = fopen(fname, "r")) == NULL)
   {
   	fprintf(stderr, "Can't open format file (%s)\n", fname);
      return NULL;
   }

   /* loop through format file, parsing into literals & fields */
   ret = create_csvfmt();	/* dummy head */
   fmt = ret;	/* most recent formatting element */
   while (fgets(buf, sizeof(buf), fp) != NULL)
   {
   	/* search for parameters */
      p = buf;
      while ((q = strchr(p, '$')) != NULL)
      { /* found dollar sign - check for $[n] */
      	if (*(q+1) == '[')
         { /* parameter */
         	if (q > p)
            { /* first, record the literal before this parameter */
            	fmtnew = create_csvfmt();	/* create new element */
               fmtnew->type = CSV_LIT; /* literal */
               /* need copy of string up to dollar sign */
               *q = '\0';	/* copy up to dollar sign */
               fmtnew->data.lit = strdup(p);
			      append_fmt(ret, fmtnew);
      		}
            /* entry created for literal, now field */
            fmtnew = create_csvfmt();
            fmtnew->type = CSV_FLD;	/* field number */
            fmtnew->data.fld = atoi(q+2); /* field number */
            for (p=q+2; isdigit(*p); p++); /* end of field number */
            if (*p == ':')
            { /* width of output string specified */
            	fmtnew->size = atoi(++p);	/* width */
               while (isdigit(*p) || *p=='-') p++;	/* skip digits */
            }
            if (*p == ']') p++;
            else fprintf(stderr, "Warning: unterminated field number (%s)\n", p);
            append_fmt(ret, fmtnew);
         }
         else
         { /* just a dollar sign - make it a literal */
         	fmtnew = create_csvfmt();
            fmtnew->type = CSV_LIT;
            /* save character after dollar sign */
            c = *(q+1);
            *(q+1) = '\0'; /* termiante literal */
            fmtnew->data.lit = strdup(p);
            *(q+1) = c; /* restore next character */
            append_fmt(ret, fmtnew);
         	p = q+1;
         }
      }

      /* handle trailing (or only) literal) */
      fmtnew = create_csvfmt();
      fmtnew->type = CSV_LIT;
      fmtnew->data.lit = strdup(p);
      append_fmt(ret, fmtnew);
   }
   return ret;
}

void append_fmt(struct csvfmt *ret, struct csvfmt *fmt)
{
	struct csvfmt *t;
   char buf[BUFSIZ];
#if EBUG
   printf("\nAppending format type %d", fmt->type);
   if (fmt->type == CSV_LIT) printf(" >>%s<<\n", fmt->data.lit);
   else printf(", fld=%d\n", fmt->data.fld);
#endif
	if (ret == NULL) return;
   for (t=ret; t->next != NULL; t = t->next) /* end of chain */
   {
#if EBUG
   	printf("Type %d", t->type);
      if (t->type == CSV_LIT) printf(" >>%s<<\n", t->data.lit);
      else printf(", fld=%d\n", t->data.fld);
#endif
   }
   t->next = fmt;
#if EBUG
   printf("Tail is type %d", t->type);
   if (t->type == CSV_LIT) printf(" >>%s<<\n", t->data.lit);
   else printf(", fld=%d\n", t->data.fld);

   printf("New tail is type %d", fmt->type);
   if (fmt->type == CSV_LIT) printf(" >>%s<<\n", fmt->data.lit);
   else printf(", fld=%d\n", fmt->data.fld);
   printf("Press ENTER to continue: ");
   gets(buf);
#endif
}

void csvfmt(FILE *fpout, struct csvfmt *fmt, int nskip, char *fname)
{ /* process CSV file and output as per format */
	FILE *fp;
   char buf[BUFSIZ];
   int csv;
   struct csvfmt *f;
   int maxfld;
   char *p;
   int i;

	if ((fp = fopen(fname, "r")) == NULL)
   {
   	fprintf(stderr, "Can't open input file (%s)\n", fname);
      return;
   }

   /* dump_fmt(fmt); */

   csv = wrpCsvCreate();	/* create CSV structure */
   for (i=0; i<nskip; i++) wrpCsvReadLine(csv, fp); /* skip lines */
   while (wrpCsvReadLine(csv, fp) == WRPCSV_OK)
   {
   	maxfld = wrpCsvGetNumFields(csv);
      /* loop through formats, outputting data for this line */
      for (f=fmt; f != NULL; f = f->next)
      {
      	if (f->type == CSV_LIT) fputs(f->data.lit, fpout);
         else if (f->type == CSV_FLD)
         {
         	if (f->data.fld >= maxfld || f->data.fld < 0) fputs(no_data, fpout);
         	else
            {
            	p = wrpCsvGetField(csv, f->data.fld);
	            if (f->size == 0) fputs(p, fpout);
               else
               { /* width of field specified */
               	sprintf(buf, "%%%d.%ds", f->size, f->size>0?f->size:-f->size);
/**************
printf("fmt >>%s<<, p >>%s<<\n", buf, p);
printf(">>%-9.9s<<\n", p);
**************/
                  fprintf(fpout, buf, p);
               }
            }
         }
      }
   }
}

struct csvfmt *create_csvfmt(void)
{
	struct csvfmt *ret;

   if ((ret = (struct csvfmt *)malloc(sizeof(*ret))) != NULL)
   {
   	memset(ret, 0, sizeof(*ret));
   }
}

void free_csvfmt(struct csvfmt *fmt)
{
	if (fmt == NULL) return;
}

void dump_fmt(struct csvfmt *fmt)
{
	char buf[BUFSIZ];

	printf("Dumping format(s) ...\n");
   while (fmt != NULL)
   {
   	printf("\nFormat type=%d, width=%d", fmt->type, fmt->size);
      if (fmt->type == CSV_LIT) printf(", lit >>%s<<\n", fmt->data.lit);
      else printf(", Field=%d\n", fmt->data.fld);
      fmt = fmt->next;
      printf("Press ENTER to continue:");
      gets(buf);
   }
}

void copy_file(FILE *fpin, FILE *fpout)
{
	int c; /* input character */

	rewind(fpin);	/* make sure we are at the start of the file */

   while ((c=fgetc(fpin)) != EOF) fputc(c, fpout); /* copy the file */

}

